// Bixorama.js
// An equirectangular 360 player using HTML5, Javascript and WebGL
//
// Copyright Outerspace Software - All rights reserved
// http://www.outerspace-software.com/
//
// You may use this code on a royalty free base for as many projects as you like when you have purchased a license of Bixorama:
// http://www.outerspace-software.com/bixorama

//Bixorama variables
var Bixorama_canvas = 0;
var Bixorama_gl = 0;

//Matrices
var Bixorama_pmatrix = 0;
var Bixorama_wmatrix = 0;

//Texture
var Bixorama_texture = 0;
var Bixorama_image = 0;
var Bixorama_video = 0;
var Bixorama_video_width = 0;
var Bixorama_video_height = 0;

//Mesh
var Bixorama_vertex = 0;
var Bixorama_uv = 0;
var Bixorama_index = 0;

var Bixorama_vertex_buffer = 0;
var Bixorama_uv_buffer = 0;
var Bixorama_index_buffer = 0;

//Program
var Bixorama_program = 0;

//Rotation and FOV
var Bixorama_heading = 0;
var Bixorama_pitch = 0;
var Bixorama_fov = 0;
var Bixorama_devicerotation = [1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1];
var Bixorama_deviceorientation = [1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1];

//Mouse
var Bixorama_mouse_down = false;
var Bixorama_mouse_x = 0;
var Bixorama_mouse_y = 0;
var Bixorama_heading_old = 0;
var Bixorama_pitch_old = 0;

//IE and Edge workaround
var Bixorama_workaround = false;
var Bixorama_workaround_canvas = 0;
var Bixorama_workaround_context = 0;

function Bixorama(inCanvasID, inURL, inHeading, inPitch, inFOV, inOrientationTracking, inLowPoly) {

    //Default parameters
    if(typeof inHeading == 'undefined') inHeading=0;
    if(typeof inPitch == 'undefined') inPitch=0;
    if(typeof inFOV == 'undefined') inFOV=90;
    if(typeof inOrientationTracking == 'undefined') inOrientationTracking=false;
    if(typeof inLowPoly == 'undefined') inLowPoly=false;


    //IE and Edge can't handle video textures
    //Turn on the workaround here
    if (Bixorama_IsVideo(inURL)) {
        if (Bixorama_IsIEorEdge()) {
            Bixorama_workaround = true;
        }
    }

    Bixorama_heading = Bixorama_D2R(inHeading);
    Bixorama_pitch = Bixorama_D2R(inPitch);
    Bixorama_fov = Bixorama_D2R(inFOV);

    Bixorama_InitWebGL(inCanvasID);
    Bixorama_OnResize();
    Bixorama_InitTexture(inURL);
    Bixorama_InitMesh(inLowPoly);
    Bixorama_InitShaders();
    Bixorama_Render();

    if ('ontouchstart' in document.documentElement) {
        //Tablet with touchscreen
        $("#" + inCanvasID).bind("touchstart", function (e) {
            Bixorama_OnTouchDown(e.originalEvent);
        });
        $("#" + inCanvasID).bind("touchmove", function (e) {
            Bixorama_OnTouchMove(e.originalEvent);
        });
        $("#" + inCanvasID).bind("touchend", function (e) {
            Bixorama_OnTouchUp(e.originalEvent);
        });
    }
    else {
        //Desktop with mouse
        $("#" + inCanvasID).mousedown(function (e) {
            Bixorama_OnMouseDown(e);
        });
        $("#" + inCanvasID).mousemove(function (e) {
            Bixorama_OnMouseMove(e);
        });
        $("#" + inCanvasID).mouseup(function (e) {
            Bixorama_OnMouseUp(e);
        });
        $("#" + inCanvasID).mouseleave(function (e) {
            Bixorama_OnMouseUp(e);
        });
    }

    window.addEventListener("resize", function () {
        Bixorama_OnResize();
    });

    window.addEventListener("keydown", function (e) {
        Bixorama_OnKeyDown(e);
    });

    if(inOrientationTracking){
	   window.addEventListener("deviceorientation", handleDeviceOrientation, true);
	   window.addEventListener("orientationchange", handleOrientationChange);
    }
}

function Bixorama_InitWebGL(inCanvasID) {
    Bixorama_canvas = document.getElementById(inCanvasID);
    Bixorama_gl = Bixorama_canvas.getContext("webgl") || Bixorama_canvas.getContext("experimental-webgl");
}

function Bixorama_GetCanvasWidth() {
    return window.innerWidth;
}

function Bixorama_GetCanvasHeight() {
    return window.innerHeight;
}

function Bixorama_InitMatrices() {
    var w = Bixorama_GetCanvasWidth();
    var h = Bixorama_GetCanvasHeight();
    Bixorama_pmatrix = Bixorama_GetProjectionMatrix(Bixorama_fov, w / h, 1, 100);
    Bixorama_wmatrix = Bixorama_GetRotationMatrix(Bixorama_heading, Bixorama_pitch);
}

function Bixorama_GetProjectionMatrix(inFOV, inAspectRatio, inZMin, inZMax) {
    var a = Math.tan(inFOV * 0.5);
    return [
    0.5 / a, 0, 0, 0,
    0, 0.5 * inAspectRatio / a, 0, 0,
    0, 0, -(inZMax + inZMin) / (inZMax - inZMin), -1.0,
    0, 0, (-2.0 * inZMax * inZMin) / (inZMax - inZMin), 0];
}

function Bixorama_GetRotationMatrix(inHeading, inPitch) {
    var sa = Math.sin(inHeading);
    var ca = Math.cos(inHeading);
    var sb = Math.sin(inPitch);
    var cb = Math.cos(inPitch);

    var matrix=[
    ca, sa * sb, sa * cb, 0,
    0, cb, -sb, 0,
    -sa, ca * sb, ca * cb, 0,
    0, 0, 0, 1];

    var M = Bixorama_RotateMatrix(Bixorama_devicerotation, Bixorama_deviceorientation);
    return Bixorama_RotateMatrix(M, matrix);
}

//Rotation order device is ZXY
function handleDeviceOrientation(event){
  var absolute = event.absolute;
  var alpha    = Bixorama_D2R(event.alpha);
  var beta     = Bixorama_D2R(event.beta);
  var gamma    = Bixorama_D2R(event.gamma);

  var MZ = [Math.cos(alpha),-Math.sin(alpha),0,0,		Math.sin(alpha),Math.cos(alpha),0,0,	0,0,1,0,				0,0,0,1];
  var MX = [1,0,0,0,						0,Math.cos(beta),-Math.sin(beta),0,	0,Math.sin(beta),Math.cos(beta),0,	0,0,0,1];
  var MY = [Math.cos(gamma),0,Math.sin(gamma),0,		0,1,0,0,				-Math.sin(gamma),0,Math.cos(gamma),0,	0,0,0,1];

  //Use device standing up
  var MA = [1,0,0,0,	0,0,1,0,	0,-1,0,0,	0,0,0,1];

  Bixorama_devicerotation=Bixorama_RotateMatrix(MA,MZ);
  Bixorama_devicerotation=Bixorama_RotateMatrix(Bixorama_devicerotation,MX);
  Bixorama_devicerotation=Bixorama_RotateMatrix(Bixorama_devicerotation,MY);

  Bixorama_InitMatrices();
}

//Correct orientation screen (-90,0,90)
function handleOrientationChange(){
  var a=-Bixorama_D2R(window.orientation);

  Bixorama_deviceorientation = [Math.cos(a),-Math.sin(a),0,0,		Math.sin(a),Math.cos(a),0,0,	0,0,1,0,				0,0,0,1];

  Bixorama_InitMatrices();
}

function Bixorama_RotateMatrix(inMatrix0, inMatrix1){
	var R = new Array();
	var c = 0;
	var r = 0;
	for (var i = 0; i < 16; i++){

		R[i] =	  inMatrix0[r]     * inMatrix1[c]
			+ inMatrix0[r + 1] * inMatrix1[c + 4]
			+ inMatrix0[r + 2] * inMatrix1[c + 8]
			+ inMatrix0[r + 3] * inMatrix1[c + 12];

		c++;
		if (c == 4){
			c = 0;
			r += 4;
		}
	}
	return R;
}

function Bixorama_InitTexture(inURL) {

    Bixorama_texture = Bixorama_gl.createTexture();

    if (Bixorama_IsVideo(inURL)) {
        Bixorama_video = document.createElement('video');
        Bixorama_video.style.visibility = "hidden";
        Bixorama_video.preload = "auto";
        Bixorama_video.loop = true;
        Bixorama_video.autoplay = true;
        Bixorama_video.src = inURL;

        Bixorama_video.addEventListener("loadedmetadata", function (e) {
            Bixorama_video_width = this.videoWidth;
            Bixorama_video_height = this.videoHeight;

            if (Bixorama_workaround && Bixorama_video_width > 0 && Bixorama_video_height > 0) {
                Bixorama_workaround_canvas = document.createElement('canvas');
                Bixorama_workaround_canvas.style.visibility = "hidden";
                Bixorama_workaround_canvas.width = Bixorama_video_width;
                Bixorama_workaround_canvas.height = Bixorama_video_height;
                Bixorama_workaround_context = Bixorama_workaround_canvas.getContext('2d');
            }

        }, false);
    }
    else {
        Bixorama_image = new Image();
        Bixorama_image.onload = function () { Bixorama_HandleTextureLoaded(); }
        Bixorama_image.src = inURL;
    }
}

function Bixorama_HandleTextureLoaded() {
    Bixorama_gl.bindTexture(Bixorama_gl.TEXTURE_2D, Bixorama_texture);
    Bixorama_gl.texImage2D(Bixorama_gl.TEXTURE_2D, 0, Bixorama_gl.RGBA, Bixorama_gl.RGBA, Bixorama_gl.UNSIGNED_BYTE, Bixorama_image);
    Bixorama_gl.generateMipmap(Bixorama_gl.TEXTURE_2D);
    Bixorama_gl.bindTexture(Bixorama_gl.TEXTURE_2D, null);
}

function Bixorama_InitMesh(inLowPoly) {
    var Radius = 50;
    var ResX = inLowPoly ? 10 : 100;
    var ResY = inLowPoly ? 5 : 50;

    Bixorama_vertex = new Array();
    Bixorama_uv = new Array;
    Bixorama_index = new Array();

    var i = 0;
    var j = 0;
    for (var y = 0; y < ResY; y++) {
        for (var x = 0; x < ResX; x++) {

            var H = -0.5*Math.PI + (x * Math.PI * 2) / (ResX - 1);
            var P = Math.PI * 0.5 + (y * Math.PI) / (ResY - 1);

            Bixorama_vertex[i++] = Radius * Math.cos(H) * Math.cos(P);
            Bixorama_vertex[i++] = Radius * Math.sin(P);
            Bixorama_vertex[i++] = Radius * Math.sin(H) * Math.cos(P);

            Bixorama_uv[j++] = (x * 1.0) / (ResX - 1);
            Bixorama_uv[j++] = (y * 1.0) / (ResY - 1);
        }
    }

    i = 0;
    for (var y = 0; y < ResY - 1; y++) {
        for (var x = 0; x < ResX - 1; x++) {

            if (y > 0) {
                Bixorama_index[i++] = y * ResX + x;
                Bixorama_index[i++] = y * ResX + x + 1;
                Bixorama_index[i++] = (y + 1) * ResX + x;
            }

            if (y < ResY - 1) {
                Bixorama_index[i++] = y * ResX + x + 1;
                Bixorama_index[i++] = (y + 1) * ResX + x + 1;
                Bixorama_index[i++] = (y + 1) * ResX + x;
            }
        }
    }

    //Create vertex buffer
    Bixorama_vertex_buffer = Bixorama_gl.createBuffer();
    Bixorama_gl.bindBuffer(Bixorama_gl.ARRAY_BUFFER, Bixorama_vertex_buffer);
    Bixorama_gl.bufferData(Bixorama_gl.ARRAY_BUFFER, new Float32Array(Bixorama_vertex), Bixorama_gl.STATIC_DRAW);

    //Create uv buffer
    Bixorama_uv_buffer = Bixorama_gl.createBuffer();
    Bixorama_gl.bindBuffer(Bixorama_gl.ARRAY_BUFFER, Bixorama_uv_buffer);
    Bixorama_gl.bufferData(Bixorama_gl.ARRAY_BUFFER, new Float32Array(Bixorama_uv), Bixorama_gl.STATIC_DRAW);

    //Create index buffer
    Bixorama_index_buffer = Bixorama_gl.createBuffer();
    Bixorama_gl.bindBuffer(Bixorama_gl.ELEMENT_ARRAY_BUFFER, Bixorama_index_buffer);
    Bixorama_gl.bufferData(Bixorama_gl.ELEMENT_ARRAY_BUFFER, new Uint16Array(Bixorama_index), Bixorama_gl.STATIC_DRAW);
}

function Bixorama_InitShaders() {

    var VCode = "attribute vec3 position;" +
    "uniform mat4 pmatrix;" +
    "uniform mat4 wmatrix;" +
    "attribute vec2 aTextureCoord;" +
    "varying highp vec2 vTextureCoord;" +
    "void main(void){" +
    "gl_Position = pmatrix*wmatrix*vec4(position, 1.0);" +
    "vTextureCoord = aTextureCoord;" +
    "}";

    var FCode = "precision mediump float;" +
    "varying highp vec2 vTextureCoord;" +
    "uniform sampler2D uSampler;" +
    "void main(void) {" +
    "gl_FragColor = texture2D(uSampler, vec2(vTextureCoord.s, vTextureCoord.t));" +
    "}";

    //Compile vertex shader
    var v_shader = Bixorama_gl.createShader(Bixorama_gl.VERTEX_SHADER);
    Bixorama_gl.shaderSource(v_shader, VCode);
    Bixorama_gl.compileShader(v_shader);

    //Compile fragment shader
    var f_Shader = Bixorama_gl.createShader(Bixorama_gl.FRAGMENT_SHADER);
    Bixorama_gl.shaderSource(f_Shader, FCode);
    Bixorama_gl.compileShader(f_Shader);

    //Create program
    Bixorama_program = Bixorama_gl.createProgram();
    Bixorama_gl.attachShader(Bixorama_program, v_shader);
    Bixorama_gl.attachShader(Bixorama_program, f_Shader);
    Bixorama_gl.linkProgram(Bixorama_program);
    Bixorama_gl.useProgram(Bixorama_program);

    //Bind vertex
    Bixorama_gl.bindBuffer(Bixorama_gl.ARRAY_BUFFER, Bixorama_vertex_buffer);
    var position = Bixorama_gl.getAttribLocation(Bixorama_program, "position");
    Bixorama_gl.vertexAttribPointer(position, 3, Bixorama_gl.FLOAT, false, 0, 0);
    Bixorama_gl.enableVertexAttribArray(position);

    //Bind UV
    Bixorama_gl.bindBuffer(Bixorama_gl.ARRAY_BUFFER, Bixorama_uv_buffer);
    var uv = Bixorama_gl.getAttribLocation(Bixorama_program, "aTextureCoord");
    Bixorama_gl.vertexAttribPointer(uv, 2, Bixorama_gl.FLOAT, false, 0, 0);
    Bixorama_gl.enableVertexAttribArray(uv);

    //Bind index
    Bixorama_gl.bindBuffer(Bixorama_gl.ELEMENT_ARRAY_BUFFER, Bixorama_index_buffer);
}

function Bixorama_Render() {

    Bixorama_gl.depthFunc(Bixorama_gl.ALWAYS);
    Bixorama_gl.clearColor(0, 1, 0, 1);

    Bixorama_gl.viewport(0.0, 0.0, Bixorama_GetCanvasWidth(), Bixorama_GetCanvasHeight());
    Bixorama_gl.clear(Bixorama_gl.COLOR_BUFFER_BIT);

    Bixorama_gl.uniformMatrix4fv(Bixorama_gl.getUniformLocation(Bixorama_program, "pmatrix"), false, Bixorama_pmatrix);
    Bixorama_gl.uniformMatrix4fv(Bixorama_gl.getUniformLocation(Bixorama_program, "wmatrix"), false, Bixorama_wmatrix);

    Bixorama_gl.activeTexture(Bixorama_gl.TEXTURE0);
    Bixorama_gl.bindTexture(Bixorama_gl.TEXTURE_2D, Bixorama_texture);
    if (Bixorama_video) {

        if (Bixorama_workaround) {
            if (Bixorama_workaround_canvas && Bixorama_workaround_context) {
                Bixorama_workaround_context.drawImage(Bixorama_video, 0, 0, Bixorama_video_width, Bixorama_video_height);
                Bixorama_gl.texImage2D(Bixorama_gl.TEXTURE_2D, 0, Bixorama_gl.RGBA, Bixorama_gl.RGBA, Bixorama_gl.UNSIGNED_BYTE, Bixorama_workaround_canvas);
            }
        }
        else {
            Bixorama_gl.texImage2D(Bixorama_gl.TEXTURE_2D, 0, Bixorama_gl.RGBA, Bixorama_gl.RGBA, Bixorama_gl.UNSIGNED_BYTE, Bixorama_video);
        }
    }
    else {
        Bixorama_gl.bindTexture(Bixorama_gl.TEXTURE_2D, Bixorama_texture);
        Bixorama_gl.uniform1i(Bixorama_gl.getUniformLocation(Bixorama_program, "uSampler"), 0);
    }

    Bixorama_gl.texParameteri(Bixorama_gl.TEXTURE_2D, Bixorama_gl.TEXTURE_MAG_FILTER, Bixorama_gl.LINEAR);
    Bixorama_gl.texParameteri(Bixorama_gl.TEXTURE_2D, Bixorama_gl.TEXTURE_MIN_FILTER, Bixorama_gl.LINEAR);
    Bixorama_gl.texParameteri(Bixorama_gl.TEXTURE_2D, Bixorama_gl.TEXTURE_WRAP_S, Bixorama_gl.CLAMP_TO_EDGE);
    Bixorama_gl.texParameteri(Bixorama_gl.TEXTURE_2D, Bixorama_gl.TEXTURE_WRAP_T, Bixorama_gl.CLAMP_TO_EDGE);

    Bixorama_gl.drawElements(Bixorama_gl.TRIANGLES, Bixorama_index.length, Bixorama_gl.UNSIGNED_SHORT, 0);

    window.requestAnimationFrame(Bixorama_Render);
}

function Bixorama_OnMouseDown(e) {
    Bixorama_mouse_down = true;
    Bixorama_mouse_x = e.offsetX === undefined ? e.originalEvent.layerX : e.offsetX;
    Bixorama_mouse_y = e.offsetY === undefined ? e.originalEvent.layerY : e.offsetY;
    Bixorama_heading_old = Bixorama_heading;
    Bixorama_pitch_old = Bixorama_pitch;
}

function Bixorama_OnMouseUp(e) {
    Bixorama_mouse_down = false;
}

function Bixorama_OnMouseMove(e) {
    if (!Bixorama_mouse_down) return;

    var dx = (e.offsetX === undefined ? e.originalEvent.layerX : e.offsetX) - Bixorama_mouse_x;
    var dy = (e.offsetY === undefined ? e.originalEvent.layerY : e.offsetY) - Bixorama_mouse_y;

    var s = 0.002;
    Bixorama_heading = Bixorama_heading_old + dx * s;
    Bixorama_pitch = Bixorama_pitch_old + dy * s;
    Bixorama_wmatrix = Bixorama_GetRotationMatrix(Bixorama_heading, Bixorama_pitch);
}

function Bixorama_OnTouchDown(e) {
    e.preventDefault();
    Bixorama_mouse_down = true;
    Bixorama_mouse_x = e.touches[0].pageX;
    Bixorama_mouse_y = e.touches[0].pageY;
    Bixorama_heading_old = Bixorama_heading;
    Bixorama_pitch_old = Bixorama_pitch;
}

function Bixorama_OnTouchUp(e) {
    Bixorama_OnMouseUp();
}

function Bixorama_OnTouchMove(e) {
    e.preventDefault();
    if (!Bixorama_mouse_down) return;

    var dx = e.touches[0].pageX - Bixorama_mouse_x;
    var dy = e.touches[0].pageY - Bixorama_mouse_y;

    var s = 0.002;
    Bixorama_heading = Bixorama_heading_old + dx * s;
    Bixorama_pitch = Bixorama_pitch_old + dy * s;
    Bixorama_wmatrix = Bixorama_GetRotationMatrix(Bixorama_heading, Bixorama_pitch);
}

function Bixorama_OnResize() {
    Bixorama_canvas.width = Bixorama_GetCanvasWidth();
    Bixorama_canvas.height = Bixorama_GetCanvasHeight();
    Bixorama_InitMatrices();
}

function Bixorama_OnKeyDown(e) {

    var s = 0.02;

    var k = e.which;
    if (k == 109) { Bixorama_fov -= s; Bixorama_InitMatrices(); }     //-
    if (k == 107) { Bixorama_fov += s; Bixorama_InitMatrices(); }     //+
    if (k == 38) { Bixorama_pitch += s; Bixorama_InitMatrices(); }    //up
    if (k == 40) { Bixorama_pitch -= s; Bixorama_InitMatrices(); }    //down
    if (k == 37) { Bixorama_heading += s; Bixorama_InitMatrices(); }  //left
    if (k == 39) { Bixorama_heading -= s; Bixorama_InitMatrices(); }  //right
    if (k == 73) { alert("Heading " + Math.round(Bixorama_R2D(Bixorama_heading)) + "\nPitch " + Math.round(Bixorama_R2D(Bixorama_pitch)) + "\nFOV " + Math.round(Bixorama_R2D(Bixorama_fov))); }	//i (info)
}

function Bixorama_D2R(a) { return a * Math.PI / 180.0; }
function Bixorama_R2D(a) { return 180.0 * a / Math.PI; }

function Bixorama_IsVideo(inURL) {
    inURL.toLowerCase();
    if (inURL.indexOf(".mp4") >= 0) return true;
    if (inURL.indexOf(".ogg") >= 0) return true;
    if (inURL.indexOf(".webm") >= 0) return true;
    return false;
}

function Bixorama_IsIEorEdge() {
    if (/MSIE 10/i.test(navigator.userAgent)) return true;
    if (/MSIE 9/i.test(navigator.userAgent) || /rv:11.0/i.test(navigator.userAgent)) return true;
    if (/Edge\/\d./i.test(navigator.userAgent)) return true;
}